@@ -70,6 +70,25 @@ int lc_channel_set_sym_key(lc_channel_t *chan, unsigned char *key, size_t len);
 /* set public channel key */
 int lc_channel_set_pub_key(lc_channel_t *chan, unsigned char *key, size_t len);
 
+/* start a thread to answer NACKs and replay messages as required */
+int lc_channel_nack_handler(lc_channel_t *chan, int n_seconds);
+
+/* similar to lc_channel_nack_handler but with an explicit NACK thread instead
+ * of the built-in one */
+int lc_channel_nack_handler_thr(lc_channel_t *chan, int n_seconds, void *(*nack_thread)(void *));
+
+/* logs an outgoing message explicitely for the NACK thread; if lc_channel_nack_handler()
+ * has not been called this is a NOP; not required if using lc_msg_send() to send messages */
+int lc_channel_nack_add_log(lc_channel_t *chan, const void *buf, size_t len, lc_seq_t seq);
+
+/* configure the channel to check for gaps in sequence number and send NACKs */
+int lc_channel_detect_gaps(lc_channel_t *chan);
+
+/* check for sequence number gaps and send NACKs if required; if lc_channel_detect_gaps()
+ * has not been called this is a NOP; not required if using lc_msg_recv() or lc_socket_listen()
+ * to receive messages */
+int lc_channel_check_seqno(lc_channel_t *chan, lc_seq_t seq);
+
 /* blocking socket recv() */
 ssize_t lc_socket_recv(lc_socket_t *sock, void *buf, size_t len, int flags);
 
