@@ -26,6 +26,7 @@
 #include <sys/socket.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <poll.h>
 
 uint32_t ctx_id = 0;
 uint32_t sock_id = 0;
@@ -249,7 +250,7 @@ static int lc_channel_keygetorset(lc_channel_t *chan, lc_key_t *key, int type, i
 	if (set) {
 		/* before storing keys mlock channel memory */
 #ifdef HAVE_LIBSODIUM
-		if (sodium_mlock(chan, sizeof chan)) return -1;
+		if (sodium_mlock(chan, sizeof *chan)) return -1;
 #endif
 		k->key = key->key;
 		k->keylen = key->keylen;
@@ -283,6 +284,112 @@ int lc_channel_set_pub_key(lc_channel_t *chan, unsigned char *key, size_t len)
 	return lc_channel_setkey(chan, &k, LC_CODE_PUBK);
 }
 
+static void lc_close_nack_logger(lc_channel_logging_t *logging)
+{
+	pthread_cancel(logging->nack_thread);
+	pthread_join(logging->nack_thread, NULL);
+	pthread_mutex_destroy(&logging->mutex);
+	while (logging->oldest) {
+		lc_packet_log_t * go = logging->oldest;
+		logging->oldest = go->next;
+		free(go);
+	}
+	free(logging);
+}
+
+static void free_gap_detect(lc_channel_gap_detect_t * gap_detect) {
+	lc_socket_close(gap_detect->sock_nack);
+	lc_channel_free(gap_detect->chan_nack);
+	free(gap_detect);
+}
+
+int lc_channel_detect_gaps(lc_channel_t *chan)
+{
+	lc_channel_gap_detect_t * gap_detect = calloc(1, sizeof(lc_channel_gap_detect_t));
+	socklen_t looplen;
+	int isloop, saveerrno;
+	if (! gap_detect) return -1;
+	gap_detect->next = 0;
+	gap_detect->chan_nack = lc_channel_sidehash(chan, (void *)SIDE_CHANNEL_NACK, strlen(SIDE_CHANNEL_NACK));
+	if (! gap_detect->chan_nack) goto error;
+	gap_detect->sock_nack = lc_socket_new(chan->ctx);
+	if (! gap_detect->sock_nack) goto error;
+	/* if the original channel had loopback we'll need that too */
+	looplen = sizeof(isloop);
+	if (getsockopt(chan->sock->sock, IPPROTO_IPV6, IPV6_MULTICAST_LOOP, &isloop, &looplen) == 0)
+		if (isloop)
+			lc_socket_loop(gap_detect->sock_nack, 1);
+	lc_channel_bind(gap_detect->sock_nack, gap_detect->chan_nack);
+	if (chan->gap_detect) free_gap_detect(chan->gap_detect);
+	chan->gap_detect = gap_detect;
+	return 0;
+error:
+	saveerrno = errno;
+	if (gap_detect->chan_nack) lc_channel_free(gap_detect->chan_nack);
+	if (gap_detect->sock_nack) lc_socket_close(gap_detect->sock_nack);
+	free(gap_detect);
+	errno = saveerrno;
+	return -1;
+}
+
+static int do_gap_detection(lc_channel_gap_detect_t *gapd, lc_seq_t seq) {
+	/* if we've never received a packet since setting this up, we'll assume packets
+	 * before this one have all arrived (no way to know) */
+	int duplicate = 0;
+	if (! gapd->next) {
+		gapd->next = seq + 1;
+		gapd->bitmap = ~(uint64_t)0;
+		return 0;
+	}
+	/* see if this is a retrasmission / arrived out of order */
+	if (seq < gapd->next) {
+		/* arrived out of order, update the bitmap if this is one of the previous
+		 * 64 packets;
+		 * don't do "if (seq >= gapd->next - 64)" in case the subtraction overflows */
+		if (seq + 64 >= gapd->next) {
+			if (gapd->bitmap & (uint64_t)1 << (seq + 64 - gapd->next))
+				duplicate = 1;
+			gapd->bitmap |= (uint64_t)1 << (seq + 64 - gapd->next);
+		}
+	}
+	else {
+		/* we'll likely need to shift the bitmap so that it covers seq - 63 to seq;
+		 * if any bit we shift out is 0 that is the last chance to send a NACK for it */
+		while (gapd->next <= seq) {
+			if (! (gapd->bitmap & ((uint64_t)1 << 63))) {
+				uint64_t miss = htobe64(gapd->next - 64);
+				lc_socket_send(gapd->sock_nack, &miss, sizeof(miss), 0);
+			}
+			gapd->bitmap <<= 1;
+			gapd->next++;
+		}
+		/* now gapd->bitmap == seq + 1 so bit 0 needs to be set: we've seen seq! */
+		gapd->bitmap |= (uint64_t)1;
+	}
+	/* either way, check if there are any other NACKs we could do with sending; to avoid
+	 * sending too many NACKs for the same packet we only check packets older than seq */
+	if (~gapd->bitmap) {
+		int bit;
+		uint64_t mask = (uint64_t)1;
+		lc_seq_t mseq = gapd->next - 1;
+		for (bit = 0; bit < 64; bit++, mask <<= 1, mseq--) {
+			if (mseq < seq && ! (gapd->bitmap & mask)) {
+				uint64_t miss = htobe64(mseq);
+				if (lc_socket_send(gapd->sock_nack, &miss, sizeof(miss), 0) == -1)
+					duplicate = -1;
+			}
+		}
+	}
+	return duplicate;
+}
+
+int lc_channel_check_seqno(lc_channel_t *chan, lc_seq_t seq)
+{
+	if (chan && chan->gap_detect)
+		return do_gap_detection(chan->gap_detect, seq);
+	else return 0;
+}
+
 void lc_channel_free(lc_channel_t * chan)
 {
 	if (!chan) return;
@@ -293,6 +400,8 @@ void lc_channel_free(lc_channel_t * chan)
 		}
 		prev = p;
 	}
+	if (chan->logging) lc_close_nack_logger(chan->logging);
+	if (chan->gap_detect) free_gap_detect(chan->gap_detect);
 	free(chan);
 }
 
@@ -448,6 +557,130 @@ ssize_t lc_socket_send(lc_socket_t *sock, const void *buf, size_t len, int flags
 	return bytes;
 }
 
+int lc_channel_nack_handler_thr(lc_channel_t *chan, int n_seconds, void *(*nack_thread)(void *))
+{
+	pthread_attr_t attr;
+	pthread_mutexattr_t mutexattr;
+	int saveerrno;
+	lc_channel_logging_t *logging = calloc(1, sizeof(lc_channel_logging_t));
+	if (! logging) return -1;
+	logging->chan_nack = lc_channel_sidehash(chan, (void *)SIDE_CHANNEL_NACK, strlen(SIDE_CHANNEL_NACK));
+	if (! logging->chan_nack) goto error;
+	logging->sock_nack = lc_socket_new(chan->ctx);
+	if (! logging->sock_nack) goto error;
+	lc_channel_bind(logging->sock_nack, logging->chan_nack);
+	lc_channel_join(logging->chan_nack);
+	logging->chan_data = chan;
+	logging->n_seconds = n_seconds;
+	if (chan->logging) lc_close_nack_logger(chan->logging);
+	chan->logging = logging;
+	pthread_mutexattr_init(&mutexattr);
+	pthread_mutex_init(&logging->mutex, &mutexattr);
+	pthread_mutexattr_destroy(&mutexattr);
+	logging->oldest = NULL;
+	logging->newest = NULL;
+	pthread_attr_init(&attr);
+	pthread_create(&logging->nack_thread, &attr, nack_thread, logging);
+	pthread_attr_destroy(&attr);
+	return 0;
+error:
+	saveerrno = errno;
+	if (logging->chan_nack) lc_channel_free(logging->chan_nack);
+	if (logging->sock_nack) lc_socket_close(logging->sock_nack);
+	free(logging);
+	errno = saveerrno;
+	return -1;
+}
+
+static void *nack_thread(void * _logging)
+{
+	lc_channel_logging_t * logging = _logging;
+	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
+	while (1) {
+		/* wait for a NACK request for up to a second; if one arrives, process it;
+		 * also go through the list of logged messages to see if anything needs
+		 * to be removed */
+		struct pollfd pfd = {
+			.fd = logging->sock_nack->sock,
+			.events = POLLIN,
+		};
+		int ok;
+		ok = poll(&pfd, 1, 1000);
+		if (ok > 0) {
+			/* process this request */
+			uint64_t nack;
+			lc_seq_t seq;
+			if (lc_socket_recv(logging->sock_nack, &nack, sizeof(nack), 0) == -1) break;
+			seq = be64toh(nack);
+			/* see if we have this to resend */
+			pthread_mutex_lock(&logging->mutex);
+			if (logging->oldest && logging->oldest->seq <= seq && seq <= logging->newest->seq) {
+				/* the message may be in the list, and we exit the critical section
+				 * now as the only thing other thread may do is to add entries after
+				 * logging->newest and we are going to make sure we don't look there;
+				 * this improves performance for the other threads if the list gets
+				 * very long */
+				lc_packet_log_t * end = logging->newest, * msg = NULL;
+				pthread_mutex_unlock(&logging->mutex);
+				/* find this message in the list */
+				msg = logging->oldest;
+				while (msg && msg->seq < seq && msg != end) msg = msg->next;
+				/* if found, resend it (no point checking return
+				 * of lc_channel_send() here, as nowhere to send
+				 * error. Just continue. */
+				if (msg && msg->seq == seq)
+					lc_channel_send(logging->chan_data, msg->data, msg->len, 0);
+			}
+			else pthread_mutex_unlock(&logging->mutex);
+		}
+		if (logging->oldest) {
+			time_t expire;
+			pthread_mutex_lock(&logging->mutex);
+			expire = time(NULL) - logging->n_seconds;
+			while (logging->oldest && logging->oldest->time < expire) {
+				lc_packet_log_t * go = logging->oldest;
+				logging->oldest = go->next;
+				if (! logging->oldest) logging->newest = NULL;
+				free(go);
+			}
+			pthread_mutex_unlock(&logging->mutex);
+		}
+	}
+	return NULL;
+}
+
+int lc_channel_nack_handler(lc_channel_t *chan, int n_seconds)
+{
+	return lc_channel_nack_handler_thr(chan, n_seconds, nack_thread);
+}
+
+static int lc_channel_logmsg(lc_channel_logging_t *logging, const void *buf, size_t len, lc_seq_t seq)
+{
+	lc_packet_log_t * msg;
+	msg = calloc(1, sizeof(lc_packet_log_t) + len);
+	if (! msg) return -1;
+	msg->next =  NULL;
+	msg->len = len;
+	msg->time = time(NULL);
+	msg->seq = seq;
+	memcpy(msg->data, buf, len);
+	pthread_mutex_lock(&logging->mutex);
+	if (logging->newest)
+		logging->newest->next = msg;
+	else
+		logging->oldest = msg;
+	logging->newest = msg;
+	pthread_mutex_unlock(&logging->mutex);
+	return 0;
+}
+
+int lc_channel_nack_add_log(lc_channel_t *chan, const void *buf, size_t len, lc_seq_t seq)
+{
+	if (chan && chan->logging)
+		return lc_channel_logmsg(chan->logging, buf, len, seq);
+	else return 0;
+}
+
 ssize_t lc_msg_sendto(int sock, const void *buf, size_t len, struct sockaddr_in6 *sa, int flags)
 {
 	return sendto(sock, buf, len, flags, (struct sockaddr *)sa, sizeof(struct sockaddr_in6));
@@ -510,6 +743,10 @@ ssize_t lc_msg_send(lc_channel_t *chan, lc_message_t *msg)
 #endif
 	}
 
+	if (chan->logging)
+		/* chan->seq could have been changed by another thread in the meantime,
+		 * so use head->seq which we know hasn't */
+		lc_channel_logmsg(chan->logging, buf, len + ohead, be64toh(head->seq));
 	bytes = lc_msg_sendto(chan->sock->sock, buf, len + ohead, sa, 0);
 	if (bytes == -1) err = errno;
 
@@ -531,6 +768,15 @@ static int lc_socket_group_joined(lc_socket_t *sock, struct in6_addr *grp)
 }
 #endif
 
+lc_channel_t *lc_channel_by_address(lc_ctx_t *lctx, struct in6_addr *addr)
+{
+	for (lc_channel_t *p = lctx->chan_list; p; p = p->next) {
+		if (!memcmp(addr,& p->sa.sin6_addr, sizeof(struct in6_addr)))
+			return p;
+	}
+	return NULL;
+}
+
 ssize_t lc_msg_recv(lc_socket_t *sock, lc_message_t *msg)
 {
 	ssize_t zi = 0;
@@ -542,6 +788,7 @@ ssize_t lc_msg_recv(lc_socket_t *sock, lc_message_t *msg)
 	struct cmsghdr *cmsg;
 	char ctl[CMSG_SPACE(sizeof(struct in6_pktinfo))];
 	lc_message_head_t head;
+	lc_channel_t *chan = NULL;
 	int opt = 1;
 
 #ifndef IPV6_MULTICAST_ALL
@@ -563,6 +810,8 @@ recv_again:
 		if (cmsg->cmsg_type == IPV6_PKTINFO) {
 			/* may not be aligned, copy */
 			memcpy(&pi, CMSG_DATA(cmsg), sizeof(struct in6_pktinfo));
+			/* if we have a channel, we can use it for gap detection */
+			chan = lc_channel_by_address(sock->ctx, &pi.ipi6_addr);
 #ifndef IPV6_MULTICAST_ALL
 			/* destination is group we haven't joined - drop it */
 			if (!lc_socket_group_joined(sock, &pi.ipi6_addr)) goto recv_again;
@@ -586,6 +835,7 @@ recv_again:
 	msg->rnd = be64toh(head.rnd);
 	msg->timestamp = be64toh(head.timestamp);
 	msg->op = head.op;
+	if (chan && chan->gap_detect) do_gap_detection(chan->gap_detect, msg->seq);
 	if (zi > (ssize_t)sizeof(lc_message_head_t)) {
 		memcpy(msg->data, buf + sizeof(lc_message_head_t), msg->len);
 	}
@@ -625,15 +875,6 @@ static void lc_op_data_handler(lc_socket_call_t *sc, lc_message_t *msg)
 	if (sc->callback_msg) sc->callback_msg(msg);
 }
 
-lc_channel_t *lc_channel_by_address(lc_ctx_t *lctx, struct in6_addr *addr)
-{
-	for (lc_channel_t *p = lctx->chan_list; p; p = p->next) {
-		if (!memcmp(addr,& p->sa.sin6_addr, sizeof(struct in6_addr)))
-			return p;
-	}
-	return NULL;
-}
-
 static ssize_t lc_socket_recvmsg_if(lc_socket_t *sock, struct msghdr *msg, int flags,
 		struct in6_pktinfo *pi)
 {
@@ -690,7 +931,7 @@ static ssize_t lc_socket_decoded(lc_socket_t *sock, struct msghdr *msg, ssize_t
 	return bytes - crypto_secretbox_MACBYTES - crypto_secretbox_NONCEBYTES;
 }
 
-ssize_t lc_socket_recvmsg_encoded(lc_socket_t *sock, struct msghdr *msg, int flags)
+static ssize_t lc_socket_recvmsg_encoded(lc_socket_t *sock, struct msghdr *msg, int flags)
 {
 	ssize_t bytes;
 	char ctl[CMSG_SPACE(sizeof(struct in6_pktinfo))];
